from visual import *

#criar a janela de controlo de variaveis e a janela da simulação
control_scene = display(width = 500, height = 500, center = (0,0.5,0), userspin = 0, userzoom = 0, title = 'Controlo de variaveis', range = 1.2)
simulation_scene = display(x = 500, width = 600, height = 600, title ='Efeito Fotoelectrico', background = color.white)
control_scene.select()

class slider:  #criar a class slider
    def __init__(self, pos=vector(0,0,0), axis=vector(0,1,0), value=None, width=0.15, min=0, max=100, color=(1,0,0)):
        pos = vector(pos)
        axis = vector(axis)
        self.min = min
        self.max = max
        self.value = value
        self.start = pos
        self.axis = axis
        self.color = color
        self.shaft = cylinder(pos = self.start, axis=axis, radius=width/4., color=(0.85,0.85,0.85))
        self.control = sphere(pos = self.start+(self.value-self.min)/(self.max-self.min)*self.axis, radius = width/2., color = color)
        self.label = label(pos = self.control.pos,text = "%0.0f" % value, opacity = 0, box = 0, line = 0)

    def getslider(self, pos=None):
        pos = vector(pos)
        value = self.min+(self.max-self.min)*(pos.y-self.start.y)/mag(self.axis)    
        self.setslider(value)
        return value

    def setslider(self, value=None):
        if value > self.max:
            value = self.max
        if value < self.min:
            value = self.min
        self.value = value
        self.control.pos = self.start+(value-self.min)/(self.max-self.min)*self.axis
        self.label.pos = self.control.pos
        self.label.text= "%0.0f" % value
        self.control.color = self.color

#definir funcões

def n_cientifica(n):
    """Muda um número para notação cientifica"""
    a = '%E' % n
    return ( a.split('E')[0].rstrip('0').rstrip('.') + 'E' + a.split('E')[1])

def esconder_metais():
    for objects in metais_caixa:
        objects.visible = False

    for objects in metais_texto:
        objects.visible = False

def fotoeletrico_calc():
    '''Efectua os cálculos para o efeito fotoeléctrico'''
    global Erem_ev, Erem, c_onda, f, Ef, Ef_ev, Ec_ev, Ec

    Erem_ev = metais_Erem[metal_escolha]
    Erem = Erem_ev / ev #Converte a energia de remoção dos electrões do metal de Eletrão-volt para Joules
    c_onda = round(sliders[1].value) * 10**-9 #Converte o valor do comprimento de onda do slider para nanómetros
    f = C/c_onda #Calcula a frequência dos fotões
    Ef = f * h #Calcula a energia dos fotões
    Ef_ev = Ef * ev #Calcula a frequência dos fotões em eV

    if Ef > Erem: #Calcula a energia cinética máxima
        Ec_ev = Ef_ev - Erem_ev
        Ec = Ec_ev / ev
    else:
        Ec_ev = 0
        Ec = 0
       
def wav2RGB(comprimento_onda):
    """Converte um valor de comprimento de onda para rgb"""
    w = int(comprimento_onda)

    # Cor
    if 380 <= w < 440:
        R = -(w - 440.) / (440. - 350.)
        G = 0.0
        B = 1.0
    elif 440 <= w < 490:
        R = 0.0
        G = (w - 440.) / (490. - 440.)
        B = 1.0
    elif 490 <= w < 510:
        R = 0.0
        G = 1.0
        B = -(w - 510.) / (510. - 490.)
    elif 510 <= w < 580:
        R = (w - 510.) / (580. - 510.)
        G = 1.0
        B = 0.0
    elif 580 <= w < 645:
        R = 1.0
        G = -(w - 645.) / (645. - 580.)
        B = 0.0
    elif 645 <= w <= 780:
        R = 1.0
        G = 0.0
        B = 0.0
    else:
        R = 0.0
        G = 0.0
        B = 0.0

    # Correcção da intensidade
    if 380 <= w < 420:
        SSS = 0.3 + 0.7*(w - 350) / (420 - 350)
    elif 420 <= w <= 700:
        SSS = 1.0
    elif 700 < w <= 780:
        SSS = 0.3 + 0.7*(780 - w) / (780 - 700)
    else:
        SSS = 0.0    

    return [round(float(SSS*R),4), round(float(SSS*G),4), round(float(SSS*B),4)]

#Constantes

C = 3 * 10**8 #Velocidade da Luz

h = 6.626 * 10**-34 #Constante de Planck

ev = 6.242 *10**18 #Converte a energia de Joules para Electrão-Volts


#Sliders
sliders = [slider(pos=(1,0,0), color=(1,0,0), min = 0, max = 100, value = 100, axis = (0,1,0)),
           slider(pos=(-1,0,0), color=(1.0, 0.8923, 0.0), min = 100, max = 850, value = 587, axis = (0,1,0))]

#Botão para efetuar os cálculos
botao = box(pos = (0,-0.2,0), size = (2/3,2/3,0.01/3), color = color.white, material=materials.plastic)
botao_l = label(pos = (0,-0.2,0), text = ('Cálculos'), box = 1, linecolor = color.red, border = 2, opacity = 0, color = color.red)

botao_linhas = [box(pos = (0,-8/15,0), size = (2/3,0.1/3,0.05/3), color = color.blue),
                box(pos = (0,2/15,0), size = (2/3,0.1/3,0.05/3), color = color.blue),
                box(pos = (-1/3,-0.2,0), size = (0.1/3,2/3,0.05/3), color = color.blue),
                box(pos = (1/3,-0.2,0), size = (0.1/3,2/3,0.05/3), color = color.blue)]

#Menú dos metais
caixa_metais = box(pos = (0,1.4,0), size = (0.5,0.25,0.025), color = (183/255,195/255,232/255))
texto_metais = label(pos = (0,1.4,0), text = 'Metais', opacity = 0, height = 18, border = 4)
caixa_proximo = box(pos = (0,0.4,0), size = (0.5,0.25,0.025), color = (183/255,195/255,232/255))
texto_proximo = label(pos = (0,0.41,0), text = 'Próximos', opacity = 0, height = 16, border = 4)

linhas = [curve(pos = ((-0.25,1.271,0.025), (0.25,1.271,0.025)), radius = 0.0025, color = (0,0,0)),
          curve(pos = ((-0.25,1.025,0.025), (0.25,1.025,0.025)), radius = 0.0025, material = materials.emissive),
          curve(pos = ((-0.25,0.775,0.025), (0.25,0.775,0.025)), radius = 0.0025, material = materials.emissive),
          curve(pos = ((-0.25,0.525,0.025), (0.25,0.525,0.025)), radius = 0.0025, material = materials.emissive)]

metais_caixa = [box(pos = (0,1.15,0), size = (0.5,0.25,0.025), material = materials.Aluminum),
                box(pos = (0,0.9,0), size = (0.5,0.25,0.025), material = materials.Copper),
                box(pos = (0,0.65,0), size = (0.5,0.25,0.025), material = materials.Platinum),
                box(pos = (0,1.15,0), size = (0.5,0.25,0.025), material = materials.Iron),
                box(pos = (0,0.9,0), size = (0.5,0.25,0.025), material = materials.Gold),
                box(pos = (0,0.65,0), size = (0.5,0.25,0.025), material = materials.Sodium),
                box(pos = (0,1.15,0), size = (0.5,0.25,0.025), material = materials.Lead),
                box(pos = (0,0.9,0), size = (0.5,0.25,0.025), material = materials.Cobalt),
                box(pos = (0,0.65,0), size = (0.5,0.25,0.025), material = materials.Zinc)]

metais_texto = [label(pos = (0,1.15,0), text = 'Alumínio', opacity = 0, height = 15, border = 4),
                label(pos = (0,0.9,0), text = 'Cobre', opacity = 0, height = 15, border = 5),
                label(pos = (0,0.65,0), text = 'Platina', opacity = 0, height = 15, border = 5),
                label(pos = (0,1.15,0), text = 'Ferro', opacity = 0, height = 15, border = 5),
                label(pos = (0,0.9,0), text = 'Ouro', opacity = 0, height = 15, border = 5),
                label(pos = (0,0.65,0), text = 'Sódio', opacity = 0, height = 15, border = 5),
                label(pos = (0,1.15,0), text = 'Chumbo', opacity = 0, height = 15, border = 5),
                label(pos = (0,0.9,0), text = 'Cobalto', opacity = 0, height = 15, border = 5),
                label(pos = (0,0.65,0), text = 'Zinco', opacity = 0, height = 15, border = 5)]

metais_Erem = {'Alumínio': 4.08, 'Cobre': 4.7, 'Platina': 6.35, 'Ferro': 4.5, 'Ouro': 5.1, 'Sódio': 2.28, 'Chumbo': 4.14, 'Cobalto': 5.0, 'Zinco': 4.3}
texturas = {'0': 'Alumínio', '1': 'Cobre', '2': 'Platina', '3': 'Ferro', '4': 'Ouro', '5': 'Sódio', '6': 'Chumbo','7': 'Cobalto','8': 'Zinco'}

#Simulação
simulation_scene.select()
chao = box (pos = (0,-12,0), size = (50,0.2,24), color = color.white, material=materials.wood)
lampada = sphere (pos = (9,16.2,0), radius = 4, color = color.yellow)
lampada_suporte = cone (pos=(9,16.2,0), radius = 8, axis = (8,6,0), color = color.blue, material=materials.plastic)
suporte = cylinder (pos = (17,22.7,0), radius = 0.5, length = 20, color = color.red, axis = (0,-1,0), material=materials.plastic)
suporte_1= box (pos = (17, 22, 0), length = 1.1, width = 1.1, color=color.red, material=materials.plastic)
luz = cylinder (pos = (9,16.2,0), axis = (-1,-1,0), length = 27, radius = 4, opacity = 0.2 , color = color.yellow)

metal = box (pos = (-8,0,0), size = (1.5,12,12), material = materials.Aluminum)
metal_suporte = box (pos = (-11.5,0,0), size = (5.5,12,12), material=materials.plastic)
sup1 = cylinder (pos = (-11,-11.9,3), axis = (0,1,0), length = 6, color = color.blue, material=materials.plastic)
sup2 = cylinder (pos = (-11,-11.9,-3), axis = (0,1,0), length = 6, color = color.blue, material=materials.plastic)

placa = box (pos = (12,0,0), size = (1.5,12,12), color = color.yellow)
placa_suporte = box (pos = (15.5,0,0), size = (5.5,12,12), material=materials.plastic)
sup3 = cylinder (pos = (15,-11.9,3), axis = (0,1,0), length = 6, color = color.blue, material=materials.plastic)
sup4 = cylinder (pos = (15,-11.9,-3), axis = (0,1,0), length = 6, color = color.blue, material=materials.plastic)

t = 0
deltat = 0.005
n = '0'
electroes = {'0':'0', '1':'0', '2':'0'}

botao_clique = 0
alarme = 0
alarme_tempo = 50
metal_escolha = 'Alumínio' #Selecciona o Alumínio como Metal padrão
menu_clique = 0
metais_fase = 0
fotoeletrico_calc()
dragobj = None

while True:
    rate(50)
    
    if control_scene.mouse.events: #Se houver alguma acção do rato
        m = control_scene.mouse.getevent() #Guardar a acção numa variável
    
        if m.click == "left" and m.pick == botao:
            botao_clique = 1
            
        elif m.click == 'left' and m.pick == caixa_metais:
            if menu_clique == 0:
                menu_clique = 1
            else:
                menu_clique = 0
                metais_fase = 0
        
        elif m.click == 'left' and m.pick == caixa_proximo:
            if metais_fase == 2:
                metais_fase = 0
            else:
                metais_fase += 1
                
        elif m.drop == "left":
            dragobj = None
            
        elif m.drag == "left":
            for index in range(len(sliders)):
                s = sliders[index]
                if m.pick is s.control:
                    pos = m.project(normal=(0,0,1))
                    dragobj = s
                    break
                
        for objects in metais_caixa:
            if m.click == 'left' and m.pick == objects:
                for index in range(len(metais_caixa)):
                    if objects == metais_caixa[index]:
                        metal_escolha = texturas[str(index)]
                        fotoeletrico_calc()
                        
                        if metal_escolha == 'Alumínio':
                            metal.material = materials.Aluminum
                        elif metal_escolha == 'Cobre':
                            metal.material = materials.Copper
                        elif metal_escolha == 'Platina':
                            metal.material = materials.Platinum
                        elif metal_escolha == 'Ferro':
                            metal.material = materials.Iron
                        elif metal_escolha == 'Ouro':
                            metal.material = materials.Gold
                        elif metal_escolha == 'Sódio':
                            metal.material = materials.Sodium
                        elif metal_escolha == 'Chumbo':
                            metal.material = materials.Lead
                        elif metal_escolha == 'Cobalto':
                            metal.material = materials.Cobalt
                        elif metal_escolha == 'Zinco':
                            metal.material = materials.Zinc
                                  
        
    newpos = control_scene.mouse.project(normal=(0,0,1))
    if dragobj and (newpos != pos): 
        pos = newpos #Mudar a posição dum slider quando este sofre acção do rato
        dragobj.getslider(pos)
        if sliders[1].value > 380 and sliders[1].value < 780: #se o comprimento de onda estiver dentro do visivel
            sliders[1].color = wav2RGB(sliders[1].value)
            luz.visible = True
            luz.color = wav2RGB(sliders[1].value)
            lampada.color = wav2RGB(sliders[1].value)
        else:
            sliders[1].color = (0.7,0.7,0.7)
            luz.visible = False
            lampada.color = (0.7,0.7,0.7)

        #Muda a opacidade/transparência da luz conforme a sua intensidade
        luz.opacity = sliders[0].value/500    
        #Efetuar os cálculos do efeito fotoeléctrico cada vez que se mude um slider
        fotoeletrico_calc()

    if (botao_clique == 1 and botao.pos.z == 0):
        botao.pos.z -= 1/6
        botao.pos.y -= 0.06
        #Apresentar os resultados do efeito fotoeléctrico
        print("\nMetal:", metal_escolha, '\nEnergia de remoção =', Erem_ev, "ev  ou ", n_cientifica(round(Erem,21)),"J",
              "\nComprimento de onda = ", n_cientifica(round(c_onda,10)), " m",
              "\nFrequência = ", n_cientifica(round(f,-11)), " Hz",
              "\nEnergia dos Fotões = ",round(Ef_ev,3), " ev  ou ", n_cientifica(round(Ef,22)), " J",
              "\nEnergia Cinética Máxima = ", round(Ec_ev,3)," ev  ou", n_cientifica(round(Ec,22)),"J")
       
        if Ec_ev > 0:
            print("Os electrões são removidos do metal com a energia máxima cinética de", round(Ec_ev,3), "ev.")
        else:
            print("Os electrões não são removidos do metal.")
        
        alarme = 1
    
    if alarme == 1:
        alarme_tempo -= 1

    if alarme_tempo == 0:
        botao.pos.z += 1/6
        botao.pos.y += 0.06
        botao_clique = 0
        alarme = 0
        alarme_tempo = 50

    if menu_clique == 0: 
        
        for objects in linhas:
            objects.color = (0,0,0)            
        
        esconder_metais()
        caixa_proximo.visible = False
        texto_proximo.visible = False

    else:

        for objects in linhas:
            objects.color = (1,1,1)

        if metais_fase == 0:
            esconder_metais()

            for objects in metais_caixa[0:3]:
                objects.visible = True

            for objects in metais_texto[0:3]:
                objects.visible = True
   
        elif metais_fase == 1:
            esconder_metais()

            for objects in metais_caixa[3:6]:
                objects.visible = True

            for objects in metais_texto[3:6]:
                objects.visible = True

        elif metais_fase == 2:
            esconder_metais()

            for objects in metais_caixa[6:9]:
                objects.visible = True

            for objects in metais_texto[6:9]:
                objects.visible = True        

        caixa_proximo.visible = True
        texto_proximo.visible = True

    electroes_prob = random.randint(0,math.ceil(60-sliders[0].value/2.5)) #calcular a probabilidade de um eletrão aparecer(este valor depende da intensidade da luz)
    if (electroes_prob == 1 and sliders[0].value > 5 and Ec > 0):
        try:
            electrao_z = random.randint(-4,4)
            electrao_y = random.randint(math.floor(-(((16-(electrao_z**2))*2)**0.5)), math.floor(((16-(electrao_z**2))*2)**0.5))
            electroes[n] = sphere(pos = (metal.x, electrao_y, electrao_z), radius = 0.2, color = color.blue) #Criar um electrão com uma posição aleatória(dentro do feixe de luz incidente no metal)
            electroes[n].velocidade = vector(random.randint(70,100)*0.12*Ec_ev,0,0) #Criar um electrão com uma velocidade aleatória(tendo em conta a energia cinética máxima)
            n = str(int(n)+1)
            electroes[str(int(n)+2)] = '0' 
        except ValueError:
            pass
        
    #Actualizar a posição dos electrões

    for i in range(1, 200):
        try:
            electroes[str(int(n)-i)].pos += electroes[str(int(n)-i)].velocidade * deltat
        except:
            pass
        
    #Eliminar o electrão se a sua posição for maior que a da parede

    for i in range(200, 0, -1):
        try:
            if (electroes[str(int(n)-i)].pos.x > placa.pos.x):
                electroes[str(int(n)-i)].visible = False
                del electroes[str(int(n)-i)]
        except:
            pass
        
    try:
        electroes[str(int(n)-201)].visible = False
        del electroes[str(int(n)-201)]
    except:
        pass

